@@ -5,7 +5,7 @@ from notebook.utils import url_path_join
 from jupyter_cogram.jupyter_cogram_serverextension.server import (
     JupyterCogramConfigHandler,
     JupyterCogramEventHandler,
-    JupyterCogramHandler,
+    AsyncJupyterCogramHandler,
     JupyterCogramTokenHandler,
     JupyterCogramVersionHandler,
     JupyterUpgradeCogramHandler,
@@ -36,7 +36,7 @@ def load_jupyter_server_extension(nb_server_app) -> None:
             for uri, handler in [
                 ("/config", JupyterCogramConfigHandler),
                 ("/event", JupyterCogramEventHandler),
-                ("/cogram", JupyterCogramHandler),
+                ("/cogram", AsyncJupyterCogramHandler),
                 ("/token", JupyterCogramTokenHandler),
                 ("/checkVersion", JupyterCogramVersionHandler),
                 ("/upgrade", JupyterUpgradeCogramHandler),
