@@ -1,3 +1,4 @@
+import asyncio
 import json
 import logging
 import os
@@ -6,11 +7,13 @@ import sys
 import urllib.parse
 from http import HTTPStatus
 from pathlib import Path
-from typing import List, Text, Optional, Dict, Any
+from typing import List, Text, Optional, Dict, Any, Tuple
 
+import aiohttp
 import requests
 from notebook.base.handlers import IPythonHandler
 from pkg_resources import parse_version
+from tornado.iostream import StreamClosedError
 
 import jupyter_cogram
 from jupyter_cogram.jupyter_cogram_serverextension.utils import (
@@ -111,7 +114,16 @@ def is_installation_up_to_date(
         return False
 
 
-def fetch_suggestion_code(
+async def cogram_api_request(
+    url: Text, payload: Optional[Dict] = None, headers: Optional[Dict] = None
+) -> Tuple[Dict[Text, Any], int]:
+    timeout = aiohttp.ClientTimeout(total=suggestions_timeout)
+    async with aiohttp.ClientSession(timeout=timeout) as session:
+        async with session.post(url, json=payload, headers=headers) as resp:
+            return await resp.json(), resp.status
+
+
+async def async_fetch_suggestion_code(
     queries,
     cell_contents: Optional[List[Text]] = None,
     auth_token: Optional[Text] = None,
@@ -125,9 +137,9 @@ def fetch_suggestion_code(
     url = urllib.parse.urljoin(url if url else backend_url, completions_enpdoint)
     logger.debug(f"Submitting /completions request to API server: {url}")
     try:
-        res = requests.post(
-            url=url,
-            json={
+        rjs, status_code = await cogram_api_request(
+            url,
+            payload={
                 "queries": queries,
                 "cell_contents": cell_contents or [],
                 "auth_token": auth_token,
@@ -136,25 +148,24 @@ def fetch_suggestion_code(
                 "prompt_presets": prompt_presets,
                 "language": language,
             },
-            # retries=1,
         )
 
-        if res.status_code >= 400:
-            if res.json() and res.json().get("error") is not None:
-                msg = res.json()["error"]
+        if status_code >= 400:
+            if rjs and rjs.get("error"):
+                msg = rjs["error"]
             else:
-                msg = res.text
+                msg = "n/a"
 
             error_dict = {
                 "status": "error",
                 "error_msg": msg,
-                "status_code": res.status_code,
+                "status_code": status_code,
             }
 
             raise ValueError(json.dumps(error_dict))
 
-        return res.json()
-    except requests.exceptions.Timeout:
+        return rjs
+    except asyncio.TimeoutError:
         logger.error("Have Timeout")
         error_dict = {
             "status": "error",
@@ -163,6 +174,8 @@ def fetch_suggestion_code(
         }
         logger.debug(f"Submitting error dict {json.dumps(error_dict)}")
         raise ValueError(json.dumps(error_dict))
+    except StreamClosedError:
+        raise
 
 
 def get_telemetry_context() -> Dict[Text, Any]:
@@ -178,11 +191,17 @@ def get_telemetry_context() -> Dict[Text, Any]:
     return TELEMETRY_CONTEXT.copy()
 
 
-class JupyterCogramHandler(IPythonHandler):
+class AsyncJupyterCogramHandler(IPythonHandler):
     def __init__(self, application, request, **kwargs):
-        super(JupyterCogramHandler, self).__init__(application, request, **kwargs)
+        super(AsyncJupyterCogramHandler, self).__init__(application, request, **kwargs)
+        self.task: Optional[asyncio.Task] = None
 
-    def post(self):
+    def on_connection_close(self):
+        if self.task:
+            self.task.cancel()
+            self.task = None
+
+    async def post(self):
         data = json.loads(self.request.body)
         queries = data.get("queries", "")
         cell_contents = data.get("cell_contents", [])
@@ -194,7 +213,8 @@ class JupyterCogramHandler(IPythonHandler):
         url = data.get("backend_url")
 
         try:
-            suggestions = fetch_suggestion_code(
+
+            suggestions_coro = async_fetch_suggestion_code(
                 queries,
                 cell_contents,
                 auth_token,
@@ -205,6 +225,13 @@ class JupyterCogramHandler(IPythonHandler):
                 language=language,
             )
 
+            self.task = asyncio.create_task(suggestions_coro)
+
+            try:
+                suggestions = await self.task
+            except asyncio.CancelledError:
+                return
+
             logger.debug(f"Have suggestions\n{json.dumps(suggestions, indent=2)}")
             if "choices" in suggestions and suggestions["choices"]:
                 message = [
@@ -217,14 +244,14 @@ class JupyterCogramHandler(IPythonHandler):
             status = "success"
             status_code = 200
         except Exception as e:
-            logger.exception("Encountered error when submitting to API server.")
+            logger.error("Encountered error when submitting to API server.")
             try:
                 error_dict = json.loads(str(e))
                 status = error_dict["status"]
                 message = error_dict["error_msg"]
                 status_code = error_dict["status_code"]
-            except Exception:
-                logger.exception("Unknown error.")
+            except Exception as e:
+                logger.error(f"Unknown error: {e}")
                 status_code = 400
                 status = "Error"
                 message = "Unknown error."
@@ -232,7 +259,7 @@ class JupyterCogramHandler(IPythonHandler):
         response = {"status": status, "message": message}
         self.set_status(status_code)
         logger.debug(f"Returning response dict {json.dumps(response, indent=2)}")
-        self.finish(json.dumps(response))
+        await self.finish(json.dumps(response))
 
 
 class JupyterCogramTokenHandler(IPythonHandler):
@@ -457,8 +484,10 @@ class JupyterCogramLaunchHandler(IPythonHandler):
         if is_first_launch:
             status = "first_launch"
             msg = (
-                "Welcome to Cogram 👋 Get started by writing comments or code. You "
-                "can trigger completions with the Tab key."
+                "Welcome to Cogram 👋 Get started by writing comments or code, "
+                "and Cogram will generate suggestions when you hit the Tab key. "
+                "Cogram will automatically generate suggestions for you if you've "
+                "enabled \"Autosuggest\" (check the settings menu: ⚙️)."
             )
         elif is_new_version:
             status = "new_version"
