@@ -129,9 +129,9 @@ def get_auth_header(key: Text) -> Dict[Text, Text]:
     return dict(Authorization=f"Basic {_auth}")
 
 
-def run_in_thread(f: Callable, args: Tuple = (), kwargs: Dict = {}) -> None:
+def run_in_thread(f: Callable, args: Tuple = (), kwargs: Optional[Dict] = None) -> None:
     logger.debug(f"Starting to run {f.__name__} in thread.")
-    t = threading.Thread(target=f, args=args, kwargs=kwargs)
+    t = threading.Thread(target=f, args=args, kwargs=kwargs or {})
     t.start()
 
 
